
__author__ = 'Antoine Guillaume antoine.guillaume45@gmail.com'
__version__ = "0.2.6"

__all__ = ['transformers', 'classifiers', 'utils', 'interpreters']

__USE_NUMBA_CACHE__ = True
__USE_NUMBA_FASTMATH__ = True
__USE_NUMBA_NOGIL__ = True
__USE_NUMBA_PARALLEL__ = True